@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* parser.c: syntax parser */
-/* (C) 2007-2008 magicant */
+/* (C) 2007-2009 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -1029,6 +1029,7 @@ void **parse_words_and_redirects(redir_T **redirlastp, bool first)
 	    redirlastp = &redir->next;
 	} else if ((word = parse_word(noalias))) {
 	    pl_add(&wordlist, word);
+	    skip_blanks_and_comment();
 	    first = false;
 	} else {
 	    break;
@@ -1071,12 +1072,7 @@ assign_T *tryparse_assignment(void)
     ensure_buffer(1);
     if (cbuf.contents[cindex] != L'(') {
 	result->a_type = A_SCALAR;
-	if (is_token_delimiter_char(cbuf.contents[cindex])) {
-	    skip_blanks_and_comment();
-	    result->a_scalar = NULL;
-	} else {
-	    result->a_scalar = parse_word(noalias);
-	}
+	result->a_scalar = parse_word(noalias);
     } else {
 	cindex++;
 	skip_to_next_token();
@@ -1086,8 +1082,8 @@ assign_T *tryparse_assignment(void)
 	    cindex++;
 	else
 	    serror(Ngt("`%ls' missing"), L")");
-	skip_blanks_and_comment();
     }
+    skip_blanks_and_comment();
     return result;
 }
 
@@ -1183,21 +1179,25 @@ reparse:
 	}
     }
     skip_blanks_and_comment();
-    if (is_command_delimiter_char(cbuf.contents[cindex])) {
-	serror(Ngt("redirect target not specified"));
-	free(result);
-	return NULL;
-    }
     if (result->rd_type != RT_HERE && result->rd_type != RT_HERERT) {
 	result->rd_filename = parse_word(globalonly);
+	if (!result->rd_filename) {
+	    serror(Ngt("redirect target not specified"));
+	    free(result);
+	    return NULL;
+	}
     } else {
-	size_t index = cindex;
 	wchar_t *endofheredoc = parse_word_as_wcs();
-	assert(index != cindex);
+	if (endofheredoc[0] == L'\0') {
+	    serror(Ngt("here-document delimiter not specified"));
+	    free(result);
+	    return NULL;
+	}
 	result->rd_hereend = endofheredoc;
 	result->rd_herecontent = NULL;
 	pl_add(&pending_heredocs, result);
     }
+    skip_blanks_and_comment();
     return result;
 
 parse_command:
@@ -1219,7 +1219,8 @@ wordunit_T *parse_word(aliastype_T type)
  * `testfunc' is a function that determines if a character is a word delimiter.
  * It must return true for L'\0'.
  * The parsing process proceeds up to an unescaped character `testfunc' returns
- * false for. It is not an error if there is no characters to be a word. */
+ * false for. It is not an error if there is no characters to be a word, in
+ * which case NULL is returned. */
 wordunit_T *parse_word_to(aliastype_T type, bool testfunc(wchar_t c))
 {
 #if YASH_ENABLE_ALIAS
@@ -1310,7 +1311,6 @@ wordunit_T *parse_word_to(aliastype_T type, bool testfunc(wchar_t c))
     if (indq)
 	serror(Ngt("double-quote not closed"));
 
-    skip_blanks_and_comment();
     return first;
 }
 
@@ -1561,7 +1561,7 @@ parse_match:
 	pe->pe_match = parse_word_to(noalias, is_slash_or_closing_brace);
     }
 
-    /* ensure_buffer(1); */
+    ensure_buffer(1);
     if (cbuf.contents[cindex] != L'/')
 	goto check_closing_paren_and_finish;
 parse_subst:
@@ -1569,7 +1569,7 @@ parse_subst:
     pe->pe_subst = parse_word_to(noalias, is_closing_brace);
 
 check_closing_paren_and_finish:
-    /* ensure_buffer(1); */
+    ensure_buffer(1);
     if (cbuf.contents[cindex] == L'}')
 	cindex++;
     else
@@ -1753,20 +1753,14 @@ fail:
 }
 
 /* Returns a word token at the current index as a newly malloced string.
- * `cindex' is advanced to the next token by `skip_blanks_and_comment'.
- * This function never returns NULL. */
+ * `cindex' proceeds to just after the word.
+ * This function never returns NULL, but may return an empty string. */
 wchar_t *parse_word_as_wcs(void)
 {
-    size_t index = cindex;
+    size_t startindex = cindex;
     wordfree(parse_word(globalonly));
-
-    /* create a copy of the word */
-    wchar_t *result = xwcsndup(cbuf.contents + index, cindex - index);
-    /* remove the trailing blanks of the word */
-    index = cindex - index;
-    while (index-- > 0 && iswblank(result[index]));
-    result[++index] = L'\0';
-    return result;
+    assert(startindex <= cindex);
+    return xwcsndup(cbuf.contents + startindex, cindex - startindex);
 }
 
 /* Parses a compound command.
@@ -1930,7 +1924,7 @@ command_T *parse_for(void)
 	skip_blanks_and_comment();
 	result->c_forwords = parse_words_and_redirects(&redirs, false);
 	if (redirs) {
-	    serror(Ngt("redirections not allowed after `in'"));
+	    serror(Ngt("redirection not allowed after `in'"));
 	    redirsfree(redirs);
 	}
     } else {
@@ -2064,6 +2058,7 @@ void **parse_case_patterns(void)
 	    serror(Ngt("invalid character `%lc' in case pattern"),
 		    (wint_t) cbuf.contents[cindex]);
 	pl_add(&wordlist, parse_word(globalonly));
+	skip_blanks_and_comment();
 	ensure_buffer(1);
 	if (cbuf.contents[cindex] == L'|') {
 	    cindex++;
